// Copyright (C) 2016 Parrot SA
//
//    Redistribution and use in source and binary forms, with or without
//    modification, are permitted provided that the following conditions
//    are met:
//    * Redistributions of source code must retain the above copyright
//      notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright
//      notice, this list of conditions and the following disclaimer in
//      the documentation and/or other materials provided with the
//      distribution.
//    * Neither the name of Parrot nor the names
//      of its contributors may be used to endorse or promote products
//      derived from this software without specific prior written
//      permission.
//
//    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
//    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
//    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
//    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
//    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
//    OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
//    AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
//    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
//    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
//    SUCH DAMAGE.
//
//  Created by Nicolas CHRISTE, <nicolas.christe@parrot.com>

import Foundation

/**
 Drone API
 */
public class Drone {

    public enum ConnectionState: Int {
        case searching, connecting, connected, disconnected
    }

    /** Name of the connected drone. */
    public private (set) var name = ""

    /** Speed for simple moves. Speed in in % of the drone maxium speed*/
    public var speed = UInt(50)

    /** Drone connection state */
    public var connectionState: ConnectionState {
        return controller.connectionState
    }

    /** Drone flying state */
    public var flyingState: FlyingState {
        return controller.flyingState
    }

    /** Lights accessory, nil if the drone doesn't have lights */
    public var lights: Lights? {
        return controller.lights
    }

    /** Gun accessory, nil if the drone doesn't have a gun */
    public var gun: Gun? {
        return controller.gun
    }

    /** Claw accessory, nil if the drone doesn't have a claw */
    public var claw: Claw? {
        return controller.claw
    }

    // TODO: make this internal !
    public var controller = DroneController()

    /**
     Constructor
     */
    public init(name: String) {
        self.name = name
    }

    /**
     Connect to the drone
     */
    public func connect() {
        controller.execute(op: .connect(name: name))
    }

    /**
     Connect a drone by name
     - parameter name: name of the drone to connect to
     */
    public func connect(name: String) {
        self.name = name
        controller.execute(op: .connect(name: name))
    }

    /**
     Take off. This function make the drone fly and wait at about 1m height.
     */
    public func takeOff() {
        controller.execute(op:.takeOff)
    }

    /**
     Land. This function make the drone land at its current position
     */
    public func land() {
        controller.execute(op: .land)
    }

    /**
     Move in a single direction during the specified time.
     - parameter direction: direction to move
     - parameter duration: duration of the move
     */
    public func move(_ direction: MoveDirection, duration: Int) {
        switch direction {
        case .forward:
            move(params: MoveParams(longitudinalSpeed:  Int(speed)), duration: duration)
        case .backward:
            move(params: MoveParams(longitudinalSpeed:  -Int(speed)), duration: duration)
        case .left:
            move(params: MoveParams(lateralSpeed: -Int(speed)), duration: duration)
        case .right:
            move(params: MoveParams(lateralSpeed: Int(speed)), duration: duration)
        case .up:
            move(params: MoveParams(verticalSpeed: Int(speed)), duration: duration)
        case .down:
            move(params: MoveParams(verticalSpeed: -Int(speed)), duration: duration)
        }
    }

    /**
     Ask the drone to turn on itself
     - parameter direction: turn direction
     - parameter angle: angle to turn in degre (0 to 180 degre)
     */
    public func turn(_ direction: TurnDirection, angle: UInt) {
        switch direction {
        case .left:
            controller.execute(op: .turn(angle: -Int(angle)))
        case .right:
            controller.execute(op: .turn(angle: Int(angle)))
        }
    }

    /**
     Move in one direction for a specific duration
     - parameter param: move parameters
     - parameter speed: the speed to move at
     - parameter duration: duration of the movement
     */
    public func move(params: MoveParams, duration: Int = -1) {
        controller.execute(op: .move(params: params, speed: duration))
    }

    public func stopMoving() {
        controller.execute(op: .stop)
    }

    /**
     Perform a flip
     - paramter direction: flip direction
     */
    public func flip(_ direction: FlipDirection) {
        controller.execute(op: .flip(direction: direction))
    }

    /**
     Take a picture
     */
    public func takePicture() {
        controller.execute(op: .takePicture)
    }
}

