// Copyright (C) 2016 Parrot SA
//
//    Redistribution and use in source and binary forms, with or without
//    modification, are permitted provided that the following conditions
//    are met:
//    * Redistributions of source code must retain the above copyright
//      notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright
//      notice, this list of conditions and the following disclaimer in
//      the documentation and/or other materials provided with the
//      distribution.
//    * Neither the name of Parrot nor the names
//      of its contributors may be used to endorse or promote products
//      derived from this software without specific prior written
//      permission.
//
//    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
//    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
//    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
//    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
//    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
//    OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
//    AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
//    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
//    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
//    SUCH DAMAGE.
//
//  Created by Nicolas CHRISTE, <nicolas.christe@parrot.com>

import Foundation

/**
 Controller delegate, to update live view
 */
public protocol DroneControllerDelegate {
    func didSetDroneName(name: String)
    func connectionStateDidChange(state: Drone.ConnectionState)
    func flyingStateDidChange(state: FlyingState)
    func lowBattery()
}

/**
 Drone Controller managing a drone
 */
public class DroneController {

    /** range to bound angle */
    private let angleRange = -180...180
    /** range to bound commnands in % */
    private let percentRange = 0...100
    /** range to bound commands in +/- % */
    private let signedPercentRange = -100...100

    /** Delegate */
    public var delegate: DroneControllerDelegate? {
        didSet {
            self.delegate?.connectionStateDidChange(state: connectionState)
            self.delegate?.flyingStateDidChange(state: flyingState)
            if batteryLow {
                self.delegate?.lowBattery()
            }
        }
    }

    /** connection state */
    public fileprivate (set) var connectionState: Drone.ConnectionState = .disconnected {
        didSet {
            self.delegate?.connectionStateDidChange(state: connectionState)
        }
    }

    /** flying state */
    public fileprivate (set) var flyingState: FlyingState = .landed {
        didSet {
            self.delegate?.flyingStateDidChange(state: flyingState)
        }
    }

    /** true if battery level is to low to flight */
    public fileprivate (set) var batteryLow = false {
        didSet {
            self.delegate?.lowBattery()
        }
    }

    /** Lights, not nil if the drone has the light accessory */
    public fileprivate (set) var lights: Lights?
    /** Claw, not nil of the drone has the Claw accessory */
    public fileprivate (set) var claw: Claw?
    /** Gun, not nil if the drone has the gun accessory */
    public fileprivate (set) var gun: Gun?


    /** Queue running drone controller */
    private let dispatchQueue = DispatchQueue(label:"DroneController", target: nil)
    /** Timer to send pcmd */
    fileprivate var pcmdTimer: DispatchSourceTimer!
    /** Command semaphore to make commands synchrone */
    fileprivate let dispatchSemaphore = DispatchSemaphore(value: 0)
    /** Ble instance */
    fileprivate var ble: DroneBle?
    /** Protocol instance */
    fileprivate var droneProtocol: DroneProtocol?
    /** Current operation. Set while waiting for operation completion event */
    fileprivate var currentOp: Operation? = nil
    /** pcmd value */
    private var pcmd = (roll:Int8(0), pitch:Int8(0), yaw:Int8(0), gaz:Int8(0))

    /**
     Operations
     */
    public enum Operation {
        case connect(name: String)
        case takeOff
        case land
        case move(params: MoveParams, speed: Int)
        case stop
        case turn(angle: Int)
        case flip(direction: FlipDirection)
        case setLights(state: Lights.State, id: UInt8?)
        case fireGun(id: UInt8)
        case openClaw(id: UInt8)
        case closeClaw(id: UInt8)
        case takePicture
    }

    /**
     Constructor
     */
    public init() {
        self.pcmdTimer = DispatchSource.makeTimerSource(queue: dispatchQueue)
        pcmdTimer.setEventHandler {
            [unowned self] in
            let flag = self.pcmd.pitch != 0 || self.pcmd.roll != 0
            self.droneProtocol?.pcmd(flag: flag, roll: self.pcmd.roll, pitch:  self.pcmd.pitch,
                                     yaw: self.pcmd.yaw, gaz: self.pcmd.gaz)
        }
    }


    /**
     Execute an operation. Block until operation is executed
     - paramerter op: operation to execute
     */
    public func execute(op: Operation) {
        currentOp = op;
        switch currentOp! {
        // Connect a drone
        case .connect(let droneName):
            self.delegate?.didSetDroneName(name: droneName)
            if ble == nil {
                connectionState = .searching
                _ = dispatchQueue.sync() {
                    ble = DroneBle(droneId: DroneId(name: droneName), queue: dispatchQueue, delegate: self)
                }
                dispatchSemaphore.wait()
            }

        // take off, complete when the flying state is flying
        case .takeOff:
            if flyingState == .landed {
                pcmd.pitch = 0
                pcmd.roll = 0
                pcmd.gaz = 0
                _ = dispatchQueue.sync() {
                    droneProtocol?.takeOff()
                }
                dispatchSemaphore.wait()
                // wait 1 more second to ensure drone is flying
                sleep(1)
            }

        // land, complete when the flying state is landed
        case .land:
            if flyingState == .flying {
                pcmd.pitch = 0
                pcmd.roll = 0
                pcmd.gaz = 0
                _ = dispatchQueue.sync() {
                    droneProtocol?.land()
                }
                dispatchSemaphore.wait()
            }

        // move, complete after duration if > 0, immediatly else
        case .move(let params, let duration):
            pcmd.pitch = Int8(params.longitudinalSpeed.clamped(into: signedPercentRange))
            pcmd.roll = Int8(params.lateralSpeed.clamped(into: signedPercentRange))
            pcmd.yaw = Int8(params.rotationSpeed.clamped(into: signedPercentRange))
            pcmd.gaz = Int8(params.verticalSpeed.clamped(into: signedPercentRange))
            if duration > 0 {
                sleep(UInt32(duration))
                pcmd.pitch = 0
                pcmd.roll = 0
                pcmd.yaw = 0
                pcmd.gaz = 0
                // let the drone stop moving
                sleep(1)
            }

        // stop moving, complete immedially
        case .stop:
            pcmd.pitch = 0
            pcmd.roll = 0
            pcmd.yaw = 0
            pcmd.gaz = 0


        // turn, complete after a fixed delay
        case .turn(let angle):
            _ = dispatchQueue.sync() {
                droneProtocol?.cap(angle: Int16(angle.clamped(into: angleRange)))
            }
            // let the drone terminate the turn
            sleep(2)

        // perform flip, complete after a fixed delay
        case .flip(let direction):
            _ = dispatchQueue.sync() {
                droneProtocol?.flip(direction: direction)
            }
            // let the drone terminate the flip
            sleep(3)

        case .takePicture:
            _ = dispatchQueue.sync() {
                droneProtocol?.takePicture()
            }
            dispatchSemaphore.wait()

        // set light accessory state, complete immedially
        case .setLights(let state, let id):
            if let id = id {
                // delos v3
                _ = dispatchQueue.sync() {
                    droneProtocol?.accessoryLightSet(id: id, state: state)
                }
            } else {
                // delos evo
                switch state {
                case .off:
                    _ = dispatchQueue.sync() {
                        droneProtocol?.headlights(left: 0, right: 0)
                    }
                case .on(let level):
                    _ = dispatchQueue.sync() {
                        let levelCap = UInt8(Int(level).clamped(into: percentRange))
                        droneProtocol?.headlights(left: levelCap, right: levelCap)
                    }
                case .blink:
                    _ = dispatchQueue.sync() {
                        droneProtocol?.headlightsBlink()
                    }
                case .oscillate:
                    _ = dispatchQueue.sync() {
                        droneProtocol?.headlightsOscillate()
                    }
                }
            }

        case .fireGun(let id):
            _ = dispatchQueue.sync() {
                droneProtocol?.accessoryGunFire(id: id)
            }
            dispatchSemaphore.wait()


        case .openClaw(let id):
            _ = dispatchQueue.sync() {
                droneProtocol?.accessoryClowOpen(id: id)
            }
            dispatchSemaphore.wait()

        case .closeClaw(let id):
            _ = dispatchQueue.sync() {
                droneProtocol?.accessoryClowClose(id: id)
            }
            dispatchSemaphore.wait()

        }
    }
}

/** Add ble delegate facet */
extension DroneController: DroneBleDelegate {
    func droneBleDidConnect() {
        connectionState = .connecting
        droneProtocol = DroneProtocol(droneBle: ble!, delegate: self)
        droneProtocol?.connect()
    }

    func droneBleDidDisconnect() {
        connectionState = .disconnected
        droneProtocol = nil
        ble = nil
    }
}


/** Add protocol delegate facet */
extension DroneController: DroneProtocolDelegate {

    func protocolDidConnect() {
        connectionState = .connected
        if let currentOp = currentOp {
            switch currentOp {
            case .connect:
                self.currentOp = nil
                dispatchSemaphore.signal()
            default:
                break
            }
        }

        pcmdTimer.scheduleRepeating(deadline: DispatchTime.now(), interval: DispatchTimeInterval.milliseconds(100))
        pcmdTimer.resume()
    }

    func protocolDidDisconnect() {
        pcmdTimer.cancel()
        if currentOp != nil {
            currentOp = nil
            dispatchSemaphore.signal()
        }
    }

    func flyingStateDidChange(state: FlyingState) {
        if let currentOp = currentOp {
            switch currentOp {
            case .takeOff:
                if state == .flying {
                    self.currentOp = nil
                    dispatchSemaphore.signal()
                }
            case .land:
                if state == .landed {
                    self.currentOp = nil
                    dispatchSemaphore.signal()
                }
            default:
                break
            }
        }
        flyingState = state
    }

    func lowBattery() {
        batteryLow = true
    }


    func pictureStateDidChange(ready: Bool) {
        if let currentOp = currentOp {
            switch currentOp {
            case .takePicture:
                if ready {
                    self.currentOp = nil
                    dispatchSemaphore.signal()
                }
            default:
                break
            }
        }
    }
    func headlightAvailable() {
        if lights == nil {
            lights = Lights(controller: self, id: nil)
        }
    }

    func accessoryLight(available: Bool, id: UInt8, state: Lights.State) {
        if available {
            if lights == nil {
                lights = Lights(controller: self, id: id)
            }
            lights?.stateDidChange(state: state)
        } else {
            lights = nil
        }
    }

    func accessoryClaw(available: Bool, id: UInt8, opened: Bool ) {
        if available {
            if claw == nil {
                claw = Claw(controller: self, id: id)
            }
            if let currentOp = currentOp {
                switch currentOp {
                case .openClaw:
                    if opened {
                        self.currentOp = nil
                        dispatchSemaphore.signal()

                    }
                case .closeClaw:
                    if !opened {
                        self.currentOp = nil
                        dispatchSemaphore.signal()

                    }
                default:
                    break
                }
            }
        } else {
            claw = nil
            if let currentOp = currentOp {
                switch currentOp {
                case .openClaw: fallthrough
                case .closeClaw:
                    self.currentOp = nil
                    dispatchSemaphore.signal()
                default:
                    break
                }
            }
        }
    }

    func accessoryGun(available: Bool, id: UInt8, ready: Bool) {
        if available {
            if gun == nil {
                gun = Gun(controller: self, id: id)
            }
            if let currentOp = currentOp, case .fireGun = currentOp {
                if ready {
                    self.currentOp = nil
                    dispatchSemaphore.signal()
                }
            }
        } else {
            gun = nil
            if let currentOp = currentOp, case .fireGun = currentOp {
                self.currentOp = nil
                dispatchSemaphore.signal()
            }
        }
    }
}


extension Int {
    func clamped(into range: CountableClosedRange<Int>) -> Int {
        return self < range.lowerBound ? range.lowerBound :
            self > range.upperBound ? range.upperBound :
        self
    }
}


