// Copyright (C) 2016 Parrot SA
//
//    Redistribution and use in source and binary forms, with or without
//    modification, are permitted provided that the following conditions
//    are met:
//    * Redistributions of source code must retain the above copyright
//      notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright
//      notice, this list of conditions and the following disclaimer in
//      the documentation and/or other materials provided with the
//      distribution.
//    * Neither the name of Parrot nor the names
//      of its contributors may be used to endorse or promote products
//      derived from this software without specific prior written
//      permission.
//
//    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
//    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
//    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
//    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
//    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
//    OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
//    AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
//    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
//    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
//    SUCH DAMAGE.
//
//  Created by Nicolas CHRISTE, <nicolas.christe@parrot.com>

import Foundation

protocol DroneProtocolDelegate {
    func protocolDidConnect()
    func protocolDidDisconnect()
    func flyingStateDidChange(state: FlyingState)
    func lowBattery()
    func pictureStateDidChange(ready: Bool)
    // evo lights
    func headlightAvailable()
    // mambo
    func accessoryLight(available: Bool, id: UInt8, state: Lights.State)
    func accessoryClaw(available: Bool, id: UInt8, opened: Bool)
    func accessoryGun(available: Bool, id: UInt8, ready: Bool)
}

/**
 Drone ble protocol implementation
 */
class DroneProtocol {

    /** data type of ack frames */
    let typeAck: UInt8 = 1
    /** data type of data without ack frames */
    let typeDataNoAck: UInt8 = 2
    /** data type of data with ack frames */
    let typeDataWithAck: UInt8 = 4

    /**
     Commands that can be sent to the drone
     */
    enum Command {
        case setDate(String)
        case setTime(String)
        case getAllSettings
        case getAllStates
        case takeOff
        case land
        case pcmd(flag: Bool, roll: Int8, pitch: Int8, yaw: Int8, gaz: Int8)
        case cap(angle: Int16)
        case flip(direction: UInt32)
        case setMaxTilt(Float)
        case setMaxVerticalSpeed(Float)
        case cutOutMode(UInt8)
        case takePicture
        // delos evo
        case headlights(left: UInt8, right: UInt8)
        case startAnimation(id: UInt32)
        // delos v3
        case usbLightControl(id: UInt8, mode: UInt32, intensity: UInt8)
        case usbClawControl(id: UInt8, action: UInt32)
        case usbGunControl(id: UInt8, action: UInt32)

        var data: Data {
            switch self {
            case .setDate(let date):
                var data = Data(forCommand:(0,4,1))
                data.append(cmdString: date)
                return data
            case .setTime(let time):
                var data = Data(forCommand:(0,4,2))
                data.append(cmdString: time)
                return data
            case .getAllSettings:
                return Data(forCommand: (0,2,0))
            case .getAllStates:
                return Data(forCommand: (0,4,0))
            case .takeOff:
                return Data(forCommand: (2,0,1))
            case .land:
                return Data(forCommand: (2,0,3))
            case .pcmd(let flag, let roll, let pitch, let yaw, let gaz):
                var data = Data(forCommand: (2,0,2))
                data.append(data: flag ? Int8(1) : Int8(0))
                data.append(data: roll)
                data.append(data: pitch)
                data.append(data: yaw)
                data.append(data: gaz)
                data.append(data: UInt32(0))
                return data
            case .cap(let angle):
                var data = Data(forCommand: (2,4,1))
                data.append(data: angle)
                return data
            case .flip(let direction):
                var data = Data(forCommand: (2,4,0))
                data.append(data: direction)
                return data
            case .setMaxTilt(let max):
                var data = Data(forCommand: (2,8,1))
                data.append(data: max)
                return data
            case .setMaxVerticalSpeed(let max):
                var data = Data(forCommand: (2,1,0))
                data.append(data: max)
                return data
            case .takePicture:
                return Data(forCommand: (2,6,1))
            case .cutOutMode(let state):
                var data = Data(forCommand: (2,10,0))
                data.append(data: state)
                return data
            case .headlights(let left, let right):
                var data = Data(forCommand: (0,22,0))
                data.append(data: left)
                data.append(data: right)
                return data
            case .startAnimation(let id):
                var data = Data(forCommand: (0,24,0))
                data.append(data: id)
                return data
            case .usbLightControl(let id, let mode, let intensity):
                var data = Data(forCommand: (2,16,0))
                data.append(data: id)
                data.append(data: mode)
                data.append(data: intensity)
                return data
            case .usbClawControl(let id, let action):
                var data = Data(forCommand: (2,16,1))
                data.append(data: id)
                data.append(data: action)
                return data
            case .usbGunControl(let id, let action):
                var data = Data(forCommand: (2,16,2))
                data.append(data: id)
                data.append(data: action)
                return data
            }
        }
    }

    /**
     Events received from the drone
     */
    enum Event {
        case allSettingsSent
        case allStateSent
        case maxTiltChanged(value: Float, min: Float, max: Float)
        case maxVerticalSpeedChanged(value: Float, min: Float, max: Float)
        case maxRotationSpeedChanged(value: Float, min: Float, max: Float)
        case flyingStateChanged(state: UInt8)
        case batteryStateChanged(level: UInt8)
        case alertStateChanged(alert: UInt8)
        case pictureStateChanged(state: UInt8)
        // delos evo
        case headlightsState(left: UInt8, right: UInt8)
        // delos v3
        case accessoryLightState(id: UInt8, state: UInt8, intensity: UInt8, listFlags: UInt8)
        case accessoryClawState(id: UInt8, state: UInt8, listFlags: UInt8)
        case accessoryGunState(id: UInt8, state: UInt8, listFlags: UInt8)

        static func from(data:Data) -> Event? {
            switch (data[2], data[3], data[4]) {
            case (0,3,0):
                return .allSettingsSent
            case (0,5,0):
                return .allStateSent
            case (0,5,1):
                return .batteryStateChanged(level: data[6])
            case (0,23,0):
                return .headlightsState(left: data[6], right: data[7])
            case (2,9,1):
                return .maxTiltChanged(value: data.readFloat(at: 6), min: data.readFloat(at: 10), max: data.readFloat(at: 14))
            case (2,5,0):
                return .maxVerticalSpeedChanged(value: data.readFloat(at: 6),
                                                min: data.readFloat(at: 10), max: data.readFloat(at: 14))
            case (2,5,1):
                return .maxRotationSpeedChanged(value: data.readFloat(at: 6),
                                                min: data.readFloat(at: 10), max: data.readFloat(at: 14))
            case (2,3,1):
                return .flyingStateChanged(state: data[6])
            case (2,3,2):
                return .alertStateChanged(alert: data[6])
            case (2,7,1):
                return .pictureStateChanged(state: data[6])
            case (2,15,0):
                return .accessoryLightState(id: data[6], state: data[7], intensity: data[11], listFlags: data[12])
            case (2,15,1):
                return .accessoryClawState(id: data[6], state: data[7], listFlags: data[11])
            case (2,15,2):
                return .accessoryGunState(id: data[6], state: data[7], listFlags: data[11])

            default:
                return nil
            }
        }
    }

    /**
     Channel to send command with ack
     */
    class CommandChannel {
        private let sender: DroneBle.Sender
        private let ack: DroneBle.Receiver
        private var seqNr: UInt8 = 0

        init (sender: DroneBle.Sender, ack: DroneBle.Receiver) {
            self.sender = sender;
            self.ack = ack
            ack.delegate = {
                [weak self] data in
                self?.didReceive(data: data)
            }
        }

        func send(command: Command) {
            seqNr = seqNr &+ 1
            var data = command.data
            data[1] = seqNr
            sender.write(data: data)
        }

        private func didReceive (data: Data) {
            // ignore ack
        }
    }

    /**
     Control channel to send piloting commands
     */
    class ControlChannel {
        private let sender: DroneBle.Sender
        private var seqNr: UInt8 = 0
        init (sender: DroneBle.Sender) {
            self.sender = sender;
        }

        func send(command: Command) {
            seqNr = seqNr &+ 1
            var data = command.data
            data[1] = seqNr
            sender.write(data: data)
        }
    }

    /**
     Channel to receive events
     */
    class EventChannel {
        private let receiver: DroneBle.Receiver
        private let ack: DroneBle.Sender?
        private let handler: (Event)->()
        private var seqNr: UInt8 = 0

        init (receiver: DroneBle.Receiver, ack: DroneBle.Sender?, handler: @escaping (Event)->()) {
            self.receiver = receiver
            self.ack = ack
            self.handler = handler
            receiver.delegate  = {
                [weak self] data in
                self?.didReceive(data: data)
            }
        }

        private func didReceive (data: Data) {
            // acknolege all received events
            if let ack = ack {
                seqNr = seqNr &+ 1
                ack.write(data: Data(ackOf: data, withSeqNr: seqNr))
            }
            if let event = Event.from(data: data) {
                handler(event)
            }
        }
    }

    let delegate: DroneProtocolDelegate
    var commandChannel: CommandChannel!
    var controlChannel: ControlChannel!
    var eventChannel: EventChannel!
    var eventChannelNoAck: EventChannel!

    init(droneBle: DroneBle, delegate: DroneProtocolDelegate) {
        self.delegate = delegate
        self.commandChannel = CommandChannel(sender: droneBle.senders[.cmdAck]!, ack: droneBle.receivers[.cmdAck]!)
        self.controlChannel = ControlChannel(sender: droneBle.senders[.cmdNoAck]!)
        self.eventChannel = EventChannel(receiver: droneBle.receivers[.evtAck]!, ack: droneBle.senders[.evtAck]!) {
            [unowned self] event in
            self.didReceiveEvent(event)
        }
        self.eventChannelNoAck = EventChannel(receiver: droneBle.receivers[.evtNoAck]!, ack: nil) {
            [unowned self] event in
            self.didReceiveEvent(event)
        }
    }

    func connect() {
        // wait for the drone to be ready
        sleep(1)
        // execute basic connection steps

        // send date/time
        let dateFormatter = DateFormatter()
        dateFormatter.timeZone = TimeZone.current
        dateFormatter.locale = Locale.current
        let now = Date()
        dateFormatter.dateFormat = "yyyy-MM-dd"
        commandChannel.send(command: .setDate(dateFormatter.string(from: now)))
        dateFormatter.dateFormat = "'T'HHmmssZZZ"
        commandChannel.send(command: .setTime(dateFormatter.string(from: now)))

        // get all settings to start connection
        commandChannel.send(command: .getAllSettings)
    }


    func disconnected() {
        delegate.protocolDidConnect()
    }

    func takeOff() {
        commandChannel.send(command: .takeOff)
    }

    func land() {
        commandChannel.send(command: .land)
    }

    func pcmd(flag: Bool, roll: Int8, pitch: Int8, yaw: Int8, gaz: Int8) {
        controlChannel.send(command: .pcmd(flag: flag, roll: roll, pitch: pitch, yaw: yaw, gaz: gaz))
    }

    func cap(angle: Int16) {
        commandChannel.send(command: .cap(angle: angle))
    }

    func flip(direction: FlipDirection) {
        commandChannel.send(command: .flip(direction: direction.rawValue))
    }

    func takePicture() {
        commandChannel.send(command: .takePicture)
    }

    /** Delos EVO lights */
    func headlights(left: UInt8, right: UInt8) {
        commandChannel.send(command: .headlights(left: left, right: right))
    }

    func headlightsBlink() {
        commandChannel.send(command: .startAnimation(id: 1))
    }

    func headlightsOscillate() {
        commandChannel.send(command: .startAnimation(id: 2))
    }

    func accessoryGunFire(id: UInt8) {
        commandChannel.send(command: .usbGunControl(id: id, action: 0))
    }

    func accessoryClowOpen(id: UInt8) {
        commandChannel.send(command: .usbClawControl(id: id, action: 0))
    }

    func accessoryClowClose(id: UInt8) {
        commandChannel.send(command: .usbClawControl(id: id, action: 1))
    }

    func accessoryLightSet(id: UInt8, state: Lights.State) {
        switch state {
        case .off:
            commandChannel.send(command: .usbLightControl(id: id, mode: 0, intensity: 0))
        case .on(let intensity):
            commandChannel.send(command: .usbLightControl(id: id, mode: 0, intensity: UInt8(intensity)))
        case .blink:
            commandChannel.send(command: .usbLightControl(id: id, mode: 1, intensity: 0))
        case .oscillate:
            commandChannel.send(command: .usbLightControl(id: id, mode: 2, intensity: 0))
        }
    }

    private func didReceiveEvent(_ event: Event) {
        switch event {
        case .allSettingsSent:
            commandChannel.send(command: .getAllStates)
        case .allStateSent:
            // set max speed
            commandChannel.send(command: .setMaxTilt(15.0))
            commandChannel.send(command: .setMaxVerticalSpeed(2.0))
            // enable cutout mode
            commandChannel.send(command: .cutOutMode(1))
            // connected !
            print("connected!")
            delegate.protocolDidConnect()

        case .flyingStateChanged(let state):
            delegate.flyingStateDidChange(state: FlyingState.from(value: state))
        case .batteryStateChanged(_):
            break
        case .headlightsState(_, _):
            // notify drone has lights
            delegate.headlightAvailable()
        // only check left level, as we always change both light at the same time
        case .alertStateChanged(let alert):
            if (alert == 3 || alert == 4) {
                delegate.lowBattery()
            }
        case .accessoryLightState(let id, let state, let intensity, let listFlags):
            // only handle 1 accessory
            if listFlags & 0x02 == 0x02 {
                let lightState: Lights.State
                switch state {
                case 0:
                    if intensity == 0 {
                        lightState = .off
                    } else {
                        lightState = .on(UInt(intensity))
                    }
                case 1:
                    lightState = .blink
                case 2:
                    lightState = .oscillate
                default:
                    lightState = .off
                }
                delegate.accessoryLight(available: true, id: id, state: lightState)
            }  else {
                delegate.accessoryLight(available: false, id: id, state: .off)
            }

        case .accessoryClawState(let id, let state, let listFlags):
            // only handle 1 accessory
            if listFlags & 0x02 == 0x02 {
                // considere claw open if state is opened or closing
                delegate.accessoryClaw(available: true, id: id, opened: state==0 || state==3)
            } else {
                delegate.accessoryClaw(available: false, id: id, opened: false)
            }

        case .accessoryGunState(let id, let state, let listFlags):
            // only handle 1 accessory
            if listFlags & 0x02 == 0x02 {
                delegate.accessoryGun(available: true, id: id, ready: state==0)
            } else {
                delegate.accessoryGun(available: false, id: id, ready: false)
            }

        case .pictureStateChanged(let state):
            delegate.pictureStateDidChange(ready: state == 0)

        default:
            break
        }
    }
}

/**
 Extension to create data of a commands and write zero terminated string
 */
extension Data {
    init(forCommand cmd:(prj: UInt8, cls: UInt8, id:UInt8)) {
        self.init(bytes:[4, 0, cmd.prj, cmd.cls, cmd.id, 0])
    }

    init(ackOf data: Data, withSeqNr seqNr: UInt8) {
        self.init(bytes:[1, seqNr, data[1]])
    }

    mutating func append(cmdString string: String) {
        self.append(Data(bytes:Array(string.utf8)))
        self.append(Data(bytes: [UInt8(0)]))
    }

    mutating func append(data: UInt8) {
        self.append(Data(bytes: UnsafePointer<UInt8>([data]), count: MemoryLayout<UInt8>.size))
    }

    mutating func append(data: Int8) {
        UnsafePointer([data]).withMemoryRebound(to: UInt8.self, capacity: MemoryLayout<Int8>.size) {
            self.append($0, count: MemoryLayout<Int8>.size)
        }
    }

    mutating func append(data: Int16) {
        UnsafePointer([data]).withMemoryRebound(to: UInt8.self, capacity: MemoryLayout<Int16>.size) {
            self.append($0, count: MemoryLayout<Int16>.size)
        }
    }

    mutating func append(data: UInt32) {
        UnsafePointer([data]).withMemoryRebound(to: UInt8.self, capacity: MemoryLayout<UInt32>.size) {
            self.append($0, count: MemoryLayout<UInt32>.size)
        }
    }

    mutating func append(data: Float) {
        UnsafePointer([data]).withMemoryRebound(to: UInt8.self, capacity: MemoryLayout<Float>.size) {
            self.append($0, count: MemoryLayout<Float>.size)
        }
   }
    
    func readFloat(at index: Index) -> Float {
        return self.subdata(in: index..<index+MemoryLayout<Float>.size).withUnsafeBytes {
            (values: UnsafePointer<Float>)->Float in
            return values.pointee
        }
    }
}
