// Copyright (C) 2016 Parrot SA
//
//    Redistribution and use in source and binary forms, with or without
//    modification, are permitted provided that the following conditions
//    are met:
//    * Redistributions of source code must retain the above copyright
//      notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright
//      notice, this list of conditions and the following disclaimer in
//      the documentation and/or other materials provided with the
//      distribution.
//    * Neither the name of Parrot nor the names
//      of its contributors may be used to endorse or promote products
//      derived from this software without specific prior written
//      permission.
//
//    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
//    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
//    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
//    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
//    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
//    OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
//    AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
//    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
//    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
//    SUCH DAMAGE.
//
//  Created by Nicolas CHRISTE, <nicolas.christe@parrot.com>

import UIKit

public class DroneViewController: UIViewController {

    fileprivate let droneLandController = DroneLandController()
    fileprivate var droneName: String = ""
    fileprivate var batteryLow = false
    fileprivate var connectionState = Drone.ConnectionState.disconnected
    fileprivate var flyingState = FlyingState.landed

    @IBOutlet weak var statusImageView: UIImageView!
    @IBOutlet weak var statusLabel: UILabel!

    public class func makeController() -> UIViewController {
        let storyboard = UIStoryboard(name: "Main", bundle: Bundle.main)
        let viewController = storyboard.instantiateViewController(withIdentifier: "DroneViewController")
        return viewController
    }

    override public func viewDidLoad() {
        super.viewDidLoad()
        updateStatus()
    }

    func clear() {
        // land the drone if it was flying
        if !droneName.isEmpty && flyingState == .flying {
            droneLandController.landNow(droneName: droneName)
        }
        // cleanup up
        didSetDroneName(name: "")
        connectionStateDidChange(state: .disconnected)
        flyingStateDidChange(state: .landed)
        batteryLow = false
    }
}


extension DroneViewController: DroneControllerDelegate {

    public func didSetDroneName(name: String) {
        droneName = name
        statusLabel.text = name
    }

    public func connectionStateDidChange(state: Drone.ConnectionState) {
        if connectionState != state {
            connectionState = state
            updateStatus()
        }
    }

    public func flyingStateDidChange(state: FlyingState) {
        flyingState = state
    }

    public func lowBattery() {
        if  !batteryLow {
            batteryLow = true
            updateStatus()
        }
    }

    func updateStatus() {
        switch connectionState {
        case .disconnected:
            statusImageView.image = UIImage(named: "bluetooth")
            statusLabel.isHidden = true
            statusImageView.alpha = 0
        case .searching:
            statusLabel.isHidden = false
            statusImageView.alpha = 0
            UIView.animate(withDuration: 1, delay: 0, options: [.autoreverse, .repeat, .allowUserInteraction],
                           animations: {
                            self.statusImageView.alpha = 1
            })
        case .connecting:
            statusImageView.layer.removeAllAnimations()
            statusImageView.alpha = 0
            UIView.animate(withDuration: 0.1, delay: 0, options: [.autoreverse, .repeat, .allowUserInteraction],
                           animations: {
                            self.statusImageView.alpha = 1
            })
        case .connected:
            statusLabel.isHidden = false
            statusImageView.layer.removeAllAnimations()
            if batteryLow {
                statusImageView.image = UIImage(named: "battery")
                statusImageView.alpha = 0
                UIView.animate(withDuration: 0.4, delay: 0, options: [.autoreverse, .repeat, .allowUserInteraction],
                               animations: {
                                self.statusImageView.alpha = 1
                })
            } else {
                statusImageView.alpha = 1
            }
        }
    }
}
