// Copyright (C) 2016 Parrot SA
//
//    Redistribution and use in source and binary forms, with or without
//    modification, are permitted provided that the following conditions
//    are met:
//    * Redistributions of source code must retain the above copyright
//      notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright
//      notice, this list of conditions and the following disclaimer in
//      the documentation and/or other materials provided with the
//      distribution.
//    * Neither the name of Parrot nor the names
//      of its contributors may be used to endorse or promote products
//      derived from this software without specific prior written
//      permission.
//
//    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
//    FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
//    COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
//    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
//    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
//    OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
//    AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
//    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
//    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
//    SUCH DAMAGE.
//
//  Created by Nicolas CHRISTE, <nicolas.christe@parrot.com>

import Foundation

import PlaygroundSupport


public var myDroneName: String {
    get {
        if case let .string(name)? = PlaygroundKeyValueStore.current["DroneName"] {
            return name
        }
        return ""
    }
    set {
        PlaygroundKeyValueStore.current["DroneName"] = .string(newValue)
    }
}

/**
 DroneControllerDelegate forwarding into to the liveview proxy
 */
public class LiveViewProxy: DroneControllerDelegate {

    let proxy = PlaygroundPage.current.liveView as? PlaygroundRemoteLiveViewProxy

    public init() {
    }

    public func didSetDroneName(name: String) {
        proxy?.send(.dictionary(
            ["event": .string("didSetDroneName"),
             "name": .string(name)]))
    }

    public func connectionStateDidChange(state: Drone.ConnectionState) {
        proxy?.send(.dictionary(
            ["event": .string("connectionStateDidChange"),
             "state": .integer(state.rawValue)]))
    }

    public func flyingStateDidChange(state: FlyingState) {
        proxy?.send(.dictionary(
            ["event": .string("flyingStateDidChange"),
             "state": .integer(state.rawValue)]))
    }

    public func lowBattery() {
        proxy?.send(.dictionary(
            ["event": .string("lowBattery")]))
    }
}

/**
 Add liveview message handler to liveview view controller
 */
extension DroneViewController: PlaygroundLiveViewMessageHandler {

    public func liveViewMessageConnectionOpened() {
    }

    public func liveViewMessageConnectionClosed() {
        // clear live view
        clear();
    }

    public func receive(_ message: PlaygroundValue) {
        guard case let .dictionary(dict) = message else {return}
        if let entry = dict["event"] {
            if case let .string(event) = entry {
                switch event {
                case "didSetDroneName":
                    if let entry = dict["name"] {
                        if case let .string(name) = entry {
                            didSetDroneName(name: name)
                        }
                    }

                case "connectionStateDidChange":
                    if let entry = dict["state"] {
                        if case let .integer(stateValue) = entry {
                            if let state = Drone.ConnectionState(rawValue: stateValue) {
                                connectionStateDidChange(state: state)
                            }
                        }
                    }

                case "flyingStateDidChange":
                    if let entry = dict["state"] {
                        if case let .integer(stateValue) = entry {
                            if let state = FlyingState(rawValue: stateValue) {
                                flyingStateDidChange(state: state)
                            }
                        }
                    }

                case "lowBattery":
                    lowBattery()

                default:
                    break
                }
            }
        }
    }
}

